package id.co.hanoman.ngdemo;

import java.util.LinkedHashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.concurrent.TimeUnit;

import javax.annotation.PostConstruct;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.data.redis.core.RedisTemplate;
import org.springframework.data.redis.core.ValueOperations;
import org.springframework.security.core.context.SecurityContextHolder;
import org.springframework.stereotype.Component;

import com.senomas.common.RateLimit;
import com.senomas.common.RateLimiterService;
import com.senomas.common.SenomasConfiguration;
import com.senomas.common.U;

@Component
public class RateLimiterServiceImpl implements RateLimiterService {
	private static final Logger log = LoggerFactory.getLogger(RateLimiterServiceImpl.class);

	@Autowired
	private RedisTemplate<String, RateLimit> loginLimits;

	@Autowired
	SenomasConfiguration config;

	Map<String, RateLimitConfigs> configs = new LinkedHashMap<>();

	@PostConstruct
	public void init() {
		log.info("CONFIG " + U.dump(config));
		if (config.getRateLimiter() != null) {
			int gcd = 0;
			int maxLength = 0;
			for (Map.Entry<String, String> cfg : config.getRateLimiter().entrySet()) {
				List<RateLimitConfig> rlc = new LinkedList<>();
				for (String sx : cfg.getValue().split("\\s+")) {
					String vx[] = sx.split("/");
					if (vx.length == 2) {
						int limit = Integer.parseInt(vx[0]);
						int length = Integer.parseInt(vx[1]);
						maxLength = Math.max(maxLength, length);
						gcd = gcd == 0 ? length : RateLimiterServiceImpl.gcd(gcd, length);
						rlc.add(new RateLimitConfig(limit, length));
					}
				}
				configs.put(cfg.getKey(), new RateLimitConfigs(gcd, maxLength, rlc));
			}
		}
		log.info("CFG "+U.dump(configs));
	}

	@Override
	public boolean acquire(String url) {
		return acquire(url, SecurityContextHolder.getContext().getAuthentication().getName());
	}

	@Override
	public boolean acquire(String url, String id) {
		if (log.isTraceEnabled())
			log.trace("acquire([url:" + url + "], [id:" + id + "])");

		RateLimitConfigs rlc = configs.get(url);
		log.info("RATE-LIMIT ["+url+"] "+U.dump(rlc));
		if (rlc != null) {
			ValueOperations<String, RateLimit> rlv = loginLimits.opsForValue();
			String uid = url + "?" + id;
			RateLimit rl = rlv.get(uid);
			try {
				if (rl == null) {
					log.info("CREATE RATE-LIMIT ["+url+"] "+rlc.getGcd()+"  "+rlc.getLength()+"  "+U.dump(rlc.getLimits()));
					rl = new RateLimit(rlc.getGcd(), rlc.getLength());
				}
				rl.add(1);
				if (rl.isOverlimit()) {
					return false;
				}
				for (RateLimitConfig limit : rlc.getLimits()) {
					if (rl.get(limit.getLength()) > limit.getLimit()) {
						rl.setOverlimit(true);
						return false;
					}
				}
			} finally {
				if (rl != null) {
					rlv.set(uid, rl, rlc.getLength(), TimeUnit.MILLISECONDS);
				}
			}
		}
		return true;
	}

	static class RateLimitConfig {
		final int limit;
		final int length;

		public RateLimitConfig(int limit, int length) {
			this.limit = limit;
			this.length = length;
		}

		public int getLimit() {
			return limit;
		}

		public int getLength() {
			return length;
		}
	}

	static class RateLimitConfigs {
		final int gcd;
		final int length;
		final List<RateLimitConfig> limits;

		public RateLimitConfigs(int gcd, int length, List<RateLimitConfig> limits) {
			this.gcd = gcd;
			this.length = length;
			this.limits = limits;
		}

		public int getGcd() {
			return gcd;
		}
		
		public int getLength() {
			return length;
		}

		public List<RateLimitConfig> getLimits() {
			return limits;
		}
	}

	public static int gcd(int a, int b) {
		while (a != 0 && b != 0) {
			int c = b;
			b = a % b;
			a = c;
		}
		return a + b;
	}

}
